package com.maroontress.testfunnel.auditor;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import javax.swing.JTable;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;

/**
   ƥȥեͥΥơ֥ǥ롣
*/
public class FunnelTable extends JTable {

    /** ¹Ԥ뤫ɤΥåΥ */
    public static final int COL_CHECK = 0;

    /** 饹̾Υ */
    public static final int COL_CLASS = 1;

    /** ᥽å̾Υ */
    public static final int COL_NAME = 2;

    /** ֤Υ */
    public static final int COL_STATUS = 3;

    /** إåʸ */
    private String[] columns = {"", "Class", "Method", "Status"};

    /** ƥȥեͥΥơ֥ǥ */
    private AbstractTableModel tableModel;

    /** ơ֥إå */
    private JTableHeader tableHeader;

    /** ƥȥեͥ */
    private Funnel funnel;

    /**
       ƥȥեͥΥơ֥ޤ

       @param funnel ƥȥեͥ
    */
    public FunnelTable(final Funnel funnel) {
	this.funnel = funnel;
	funnel.addFunnelListener(new FunnelListener() {
	    public void recipeChanged() {
		tableModel.fireTableDataChanged();
	    }
	    public void selectionChanged() {
		tableModel.fireTableDataChanged();
	    }
	    public void statusChanged(final Test t) {
		tableModel.fireTableDataChanged();
	    }
	    public void filterChanged() {
		tableModel.fireTableDataChanged();
	    }
	    public void sortingChanged() {
		tableModel.fireTableDataChanged();
	    }
	});
	tableModel = new AbstractTableModel() {
            public String getColumnName(final int col) {
		return columns[col];
	    }
            public int getColumnCount() {
		return columns.length;
	    }
            public int getRowCount() {
		return funnel.getTestSize();
	    }
            public Object getValueAt(final int row, final int col) {
		Test t = funnel.getTest(row);
		switch (col) {
		case COL_CHECK:
		    return t.getSelected();
		case COL_CLASS:
		    return t.getClassName();
		case COL_NAME:
		    return t.getMethodName();
		case COL_STATUS:
		    //return t.getStatus();
		    return t.getStatusString();
		default:
		    break;
		}
		return null;
	    }
            public Class getColumnClass(final int col) {
		switch (col) {
		case COL_CHECK:
		    return Boolean.class;
		case COL_CLASS:
		case COL_NAME:
		    return String.class;
		case COL_STATUS:
		    //return Integer.class;
		    return String.class;
		default:
		    break;
		}
		return null;
	    }
            public boolean isCellEditable(final int row, final int col) {
		return (col == COL_CHECK);
	    }
            public void setValueAt(final Object obj,
				   final int row, final int col) {
		Test t = funnel.getTest(row);
		switch (col) {
		case COL_CHECK:
		    t.setSelected((Boolean) obj);
		    break;
		default:
		    break;
		}
	    }
	};
	setModel(tableModel);
	setColumnSelectionAllowed(false);
	setAutoResizeMode(JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS);

	getColumn("").setMaxWidth(0);
	packClassColumn();
	packStatusColumn();

	tableHeader = getTableHeader();
	tableHeader.addMouseListener(new MouseListener() {
	    public void mouseClicked(final MouseEvent ev) {
		if (ev.getClickCount() != 2) {
		    return;
		}
		sortColumn(tableHeader.columnAtPoint(ev.getPoint()));
	    }
	    public void mouseEntered(final MouseEvent ev) {
	    }
	    public void mouseExited(final MouseEvent ev) {
	    }
	    public void mousePressed(final MouseEvent ev) {
	    }
	    public void mouseReleased(final MouseEvent ev) {
	    }
	});
    }

    /**
       ־֤Υפκͤꤷޤ
    */
    private void packStatusColumn() {
	int col = COL_STATUS;
	String id = "Status";
	TableCellRenderer r = getCellRenderer(0, col);
	Component c = r.getTableCellRendererComponent(this, id, false, false,
						      0, col);
	Dimension d = c.getPreferredSize();
	TableColumn tc = getColumn(id);
	int maxWidth = d.width;

	for (int s = Test.STATUS_MIN; s <= Test.STATUS_MAX; ++s) {
	    c = r.getTableCellRendererComponent(this, Test.getStatusString(s),
						false, false, 0, col);
	    d = c.getPreferredSize();
	    if (maxWidth < d.width) {
		maxWidth = d.width;
	    }
	}

	maxWidth += 4;
	tc.setMaxWidth(maxWidth);
	tc.sizeWidthToFit();
    }

    /**
       ֥饹̾Υפκͤꤷޤ
    */
    private void packClassColumn() {
	int col = COL_CLASS;
	String id = "Class";
	TableCellRenderer r = getCellRenderer(0, col);
	Component c = r.getTableCellRendererComponent(this, id, false, false,
						      0, col);
	Dimension d = c.getPreferredSize();
	int maxWidth = d.width;
	int n = funnel.getTestSize();

	for (int row = 0; row < n; ++row) {
	    Test t = funnel.getTest(row);
	    c = r.getTableCellRendererComponent(this, t.getClassName(),
						false, false, 0, col);
	    d = c.getPreferredSize();
	    if (maxWidth < d.width) {
		maxWidth = d.width;
	    }
	}
	TableColumn tc = getColumn(id);
	maxWidth += 4;
	tc.setMaxWidth(maxWidth);
	tc.setPreferredWidth(maxWidth);
    }

    /**
       ơ֥Υ򥽡Ȥޤ

       @param col Ȥ륫
    */
    private void sortColumn(final int col) {
	switch (col) {
	case COL_STATUS:
	    funnel.sortByStatus();
	    break;
	case COL_CLASS:
	    funnel.sortByClass();
	    break;
	case COL_NAME:
	    funnel.sortByName();
	    break;
	default:
	    return;
	}
    }
}
