#include <stdlib.h>
#include "impl/Stack.h"
#include "impl/ListStack.h"

static struct StackNode *
StackNode_create(struct StackCell *c)
{
    struct StackNode *n;
    
    if ((n = (struct StackNode *)malloc(sizeof(struct StackNode))) == NULL)
        return NULL;
    n->next = NULL;
    n->value = c;
    return n;
}

static void
StackNode_free(struct StackNode *n)
{
    free(n);
}

static void
ListStack_free(struct Stack *s)
{
    struct StackNode *n, *next;
    
    for (n = s->top.node; n != NULL; n = next) {
        next = n->next;
        StackNode_free(n);
    }
    free(s);
}

static int
ListStack_push(struct Stack *s, struct StackCell *c)
{
    struct StackNode *n;

    if ((n = StackNode_create(c)) == NULL) {
	return STACK_ERROR;
    }
    n->next = s->top.node;
    s->top.node = n;
    ++(s->nodes);
    return STACK_SUCCESS;
}

static struct StackCell *
ListStack_pop(struct Stack *s)
{
    struct StackCell *value;
    struct StackNode *top;

    if (s->top.node == NULL)
        return NULL;
    value = s->top.node->value;
    top = s->top.node->next;
    StackNode_free(s->top.node);
    s->top.node = top;
    --(s->nodes);
    return value;
}

static struct StackCell *
ListStack_top(struct Stack *s)
{
    return ((s)->top.node == NULL) ? NULL : (s)->top.node->value;
}

struct Stack *
impl_ListStack_create(void)
{
    struct Stack *s;

    if ((s = (struct Stack *)malloc(sizeof(struct Stack))) == NULL)
        return NULL;
    s->top.node = NULL;
    s->nodes = 0;
    s->capacity = 0;
    s->free = ListStack_free;
    s->push = ListStack_push;
    s->pop = ListStack_pop;
    s->peek = ListStack_top;
    return s;
}
