#define com_maroontress_libld_Garbages_IMPORT
#define com_maroontress_libld_Trace_IMPORT
#define com_maroontress_libld_collector_Element_IMPORT
#define std_util_LinkedHeap_IMPORT

#include <sys/types.h>
#include <stdlib.h>
#include <com/maroontress/libld/Garbages.h>
#include <com/maroontress/libld/Trace.h>
#include <com/maroontress/libld/collector/Element.h>
#include <std/util/LinkedHeap.h>

struct Garbages {
    union LinkedHeap heap;
    struct Element *lastElement;
};

static void
freeElement(void *e)
{
    Element_free(e);
}

void
Garbages_free(struct Garbages *g)
{
    if (g == NULL)
	return;
    if (g->heap.impl != NULL) {
	LinkedHeap_forEach(g->heap, freeElement);
	LinkedHeap_free(g->heap);
    }
    free(g);
}

static int
comparator(void *e1, void *e2)
{
    return Element_compare(e1, e2);
}

struct Garbages *
Garbages_create(void)
{
    struct Garbages *g;

    if ((g = (struct Garbages *)malloc(sizeof(*g))) == NULL)
	return NULL;
    g->heap.impl = NULL;
    g->lastElement = NULL;

    if ((g->heap = LinkedHeap_create(comparator)).impl == NULL) {
	Garbages_free(g);
	return NULL;
    }
    return g;
}

int
Garbages_put(struct Garbages *g, const void *data, size_t size)
{
    void *c;

    if ((c = Element_create(data, size)) == NULL) {
	return -1;
    }
    if (LinkedHeap_put(g->heap, c) < 0) {
	Element_free(c);
 	return -1;
    }
    return 0;
}

union Trace *
Garbages_getNextTrace(struct Garbages *g)
{
    union Trace *c;

    Element_free(g->lastElement);
    if ((g->lastElement = LinkedHeap_getFirst(g->heap)) == NULL) {
	return NULL;
    }
    c = Element_getTrace(g->lastElement);
    c->type = Trace_ALLOCATED;
    return c;
}

size_t
Garbages_getSize(struct Garbages *g)
{
    return LinkedHeap_getSize(g->heap);
}
