#define com_maroontress_libld_Collector_IMPORT
#define com_maroontress_libld_Garbages_IMPORT
#define com_maroontress_libld_Trace_IMPORT
#define com_maroontress_libld_trace_AllocatedTrace_IMPORT
#define com_maroontress_testfunnel_funnel_LeakDetector_IMPORT
#define com_maroontress_testfunnel_funnel_VMATranslator_IMPORT
#define std_lang_Error_IMPORT
#define std_io_BufferedInputTube_IMPORT
#define std_io_File_IMPORT
#define std_io_InputTube_IMPORT

#include <stdio.h>
#include <stdlib.h>
#include <err.h>
#include <com/maroontress/libld/Collector.h>
#include <com/maroontress/libld/Garbages.h>
#include <com/maroontress/libld/Trace.h>
#include <com/maroontress/libld/trace/AllocatedTrace.h>
#include <com/maroontress/testfunnel/funnel/LeakDetector.h>
#include <com/maroontress/testfunnel/funnel/VMATranslator.h>
#include <std/lang/Error.h>
#include <std/io/BufferedInputTube.h>
#include <std/io/File.h>
#include <std/io/InputTube.h>

struct LeakDetector {
    union BufferedInputTube in;
    union File file;
    struct Collector *col;
    struct Garbages *g;
};

void
LeakDetector_free(struct LeakDetector *ld)
{
    if (ld == NULL)
	return;
    if (ld->in.impl == NULL)
	InputTube_close(ld->in.super.inputTube);
    File_free(ld->file);
    Collector_free(ld->col);
    Garbages_free(ld->g);
    free(ld);
}

struct LeakDetector *
LeakDetector_create(const char *traceFile)
{
    struct LeakDetector *ld;

    if ((ld = malloc(sizeof(*ld))) == NULL) {
	warn("malloc");
	return NULL;
    }
    ld->file.impl = NULL;
    ld->in.impl = NULL;
    ld->col = NULL;
    ld->g = NULL;

    if ((ld->file = File_create(traceFile)).impl == NULL
	|| (ld->in = BufferedInputTube_create(
		File_createInputTube(ld->file))).impl == NULL
	|| (ld->col = Collector_create()) == NULL
	|| (ld->g = Collector_load(ld->col, ld->in.super.inputTube)) == NULL) {
	if (ld->col != NULL) {
	    warnx("Collector_load: %s",
		  Error_toString(Collector_getLastError(ld->col)));
	}
	LeakDetector_free(ld);
	return NULL;
    }
    return ld;
}

static void
printAllocationTrace(struct AllocatedTrace *c, struct VMATranslator *x)
{
    unsigned int k;

    fprintf(stderr, "\nallocate: [pid=%u tid=0x%x] 0x%x (%d bytes)\n",
	    c->pid, c->tid, c->ptr, c->len);
    for (k = 0; k < c->count && c->stack[k] != 0; ++k) {
	VMATranslator_printReturnAddress(x, c->stack[k]);
    }
}

static void
printLeaks(struct LeakDetector *ld)
{
    struct VMATranslator *x;
    union Trace *c;

    if ((x = VMATranslator_create()) == NULL) {
	warnx("VMATranslator_create() failed.");
	return;
    }
    while ((c = Garbages_getNextTrace(ld->g)) != NULL) {
	if (c->type != Trace_ALLOCATED) {
	    continue;
	}
	printAllocationTrace(&c->allocated, x);
    }
    VMATranslator_free(x);
}

size_t
LeakDetector_printLeaks(struct LeakDetector *ld)
{
    size_t num;

    if ((num = Garbages_getSize(ld->g)) == 0)
	return 0;
    printLeaks(ld);
    fprintf(stderr, "\n%u memory leak(s) found.\n", num);
    return num;
}
