#define com_maroontress_testfunnel_funnel_VMATranslator_IMPORT
#define com_maroontress_testfunnel_funnel_vma_Place_IMPORT
#define std_db_BTree_IMPORT
#define std_lang_Chunk_IMPORT
#define std_lang_ImmutableChunk_IMPORT

#include <fcntl.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <err.h>
#include <com/maroontress/testfunnel/funnel/VMATranslator.h>
#include <com/maroontress/testfunnel/funnel/vma/Place.h>
#include <std/db/BTree.h>
#include <std/lang/Chunk.h>
#include <std/lang/ImmutableChunk.h>

static void
ImmutableChunk_set(struct ImmutableChunk *ic, const void *data, size_t size)
{
    ic->data = data;
    ic->size = size;
}

struct VMATranslator {
    union BTree db;
};

void
VMATranslator_staticInitialize(void)
{
    Place_staticInitialize();
}

struct VMATranslator *
VMATranslator_create(void)
{
    static const int MODE = 0666;
    static const int FLAGS = (O_CREAT | O_RDWR);
    struct VMATranslator *x;
    
    if ((x = (struct VMATranslator *)malloc(sizeof(*x))) == NULL) {
	warn("malloc");
	return NULL;
    }
    x->db.impl = NULL;

    if ((x->db = BTree_create(NULL, FLAGS, MODE)).impl == NULL) {
        VMATranslator_free(x);
        return NULL;
    }
    return x;
}

static void
finalizeDB(union BTree db)
{
    struct Chunk key, val;
    struct Place *place;
    int s;

    for (s = BTree_seq(db, &key, &val, BTree_FIRST); s == 0;
         s = BTree_seq(db, &key, &val, BTree_NEXT)) {
	memcpy(&place, val.data, sizeof(place));
	Place_free(place);
    }
}

void
VMATranslator_free(struct VMATranslator *x)
{
    if (x == NULL)
        return;
    if (x->db.impl != NULL) {
	finalizeDB(x->db);
        BTree_close(x->db);
    }
    free(x);
}

static struct Place *
getCachedPlace(struct VMATranslator *x, intptr_t vma)
{
    struct ImmutableChunk key, val;
    struct Chunk v;
    struct Place *place;

    ImmutableChunk_set(&key, &vma, sizeof(vma));
    if (BTree_get(x->db, key, &v) == 0) {
	memcpy(&place, v.data, sizeof(place));
	return place;
    }
    
    if ((place = Place_create(vma)) == NULL)
	return NULL;
    ImmutableChunk_set(&val, &place, sizeof(place));
    if (BTree_putNoOverwrite(x->db, key, val) != 0) {
	Place_free(place);
	return NULL;
    }
    return place;
}

void
VMATranslator_printReturnAddress(struct VMATranslator *x, intptr_t vma)
{
    struct Place *p;

    if ((p = getCachedPlace(x, vma - 1)) == NULL) {
        fprintf(stderr, "?:?: in function 0x%x\n", vma);
    }
    else {
	fprintf(stderr, "%s:%d: in funciton `%s`\n",
		Place_getFile(p), Place_getLine(p), Place_getFunction(p));
    }
}
