#define com_maroontress_testfunnel_funnel_vma_Resolver_IMPORT

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <err.h>
#include <bfd.h>
#include <com/maroontress/testfunnel/funnel/vma/Resolver.h>

struct Resolver {
    bfd *fp;
    asymbol **symbols;
    bfd_vma vma;
    const char *srcfile;
    const char *function;
    unsigned int line;
};

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

static int
initialize(struct Resolver *x, const char *aOut)
{
    char **matching;
    long n;
    unsigned int s;

    if ((x->fp = bfd_openr(aOut, NULL)) == NULL) {
	warnx("%s: bfd_openr() failed", aOut);
	return 1;
    }
    if (bfd_check_format(x->fp, bfd_archive)) {
	warnx("%s: cannot get addresses from archive", aOut);
	return 1;
    }
    if (!bfd_check_format_matches(x->fp, bfd_object, &matching)) {
	warnx("%s: bfd_check_format_matches() failed", aOut);
	return 1;
    }
    free(matching);
    if ((bfd_get_file_flags(x->fp) & HAS_SYMS) == 0) {
	warnx("%s: symbol not found", aOut);
	return 1;
    }
    if ((n = bfd_read_minisymbols(x->fp, FALSE, (PTR)&x->symbols, &s)) == 0) {
	free(x->symbols);
	x->symbols = NULL;
	n = bfd_read_minisymbols(x->fp, TRUE, (PTR)&x->symbols, &s);
    }
    if (n < 0) {
	warnx("%s: bfd_read_minisymbols() failed", aOut);
	return 1;
    }
    return 0;
}

void
Resolver_free(struct Resolver *x)
{
    free(x->symbols);
    if (x->fp != NULL)
	bfd_close(x->fp);
    free(x);
}

void
Resolver_staticInitialize(void)
{
    bfd_init();
}

struct Resolver *
Resolver_create(const char *aOut)
{
    struct Resolver *x;

    if ((x = (struct Resolver *)malloc(sizeof(*x))) == NULL) {
	warn("malloc");
	return NULL;
    }
    x->fp = NULL;
    x->symbols = NULL;

    if (initialize(x, aOut)) {
	Resolver_free(x);
	return NULL;
    }
    return x;
}

static void
finder(bfd *fp, asection *section, PTR data)
{
    struct Resolver *x = (struct Resolver *)data;
    bfd_size_type size;
    bfd_vma vma;

    if ((vma = x->vma) == 0)
	return;
    x->srcfile = NULL;
    x->function = NULL;
    x->line = 0;

    if ((bfd_get_section_flags(fp, section) & SEC_ALLOC) == 0
	|| (int)(vma -= bfd_get_section_vma(fp, section)) < 0
	/*
	  δؿGNU BinutilsΥСˤä̾㤦

	  - bfd_get_section_size_before_reloc()
	  - bfd_get_section_size()

	  ޥbfd_section_size()ϥСˤäѤʤΤǡ
	  Ѥ롣
	*/
	|| (size = bfd_section_size(fp, section)) <= vma) {
	return;
    }
    if (bfd_find_nearest_line(fp, section, x->symbols, vma,
			      &x->srcfile, &x->function, &x->line)) {
	x->vma = 0;
    }
}

int
Resolver_query(struct Resolver *x, intptr_t vma,
	       const char **srcfile, const char **function, int *line)
{
    if ((x->vma = vma) == 0)
	return 1;
    bfd_map_over_sections(x->fp, finder, x);
    *srcfile = x->srcfile;
    *function = x->function;
    *line = x->line;
    return 0;
}
